/*
Shadow config
	by Lach
	& SMS Alfredo
*/

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			freeslot(item)
		end
	end
end

function A_SetPAnim(mo, var1, var2)
	if not mo.player return end
	mo.player.panim = var2
end

function A_ModifyPFlags(mo, var1, var2) // var1 flags are added, var2 flags are removed
	local player = mo.player
	if not player return end
	player.pflags = $ & ~var2 | var1
end

function A_SpearExplode(mo)
	if not (mo.flags2 & MF2_DONTRESPAWN) // P_ExplodeMissile was called on this spear as it was spawned, fuck that
		P_RemoveMobj(mo)
		return
	end
	
	if not (mo.flags2 & MF2_AMBUSH)
		mo.scale = $ >> 1
	end
	A_Explode(mo, 0, 0)
	
	if mo.flags2 & MF2_SLIDEPUSH // spear of light
		mo.state = mo.info.xdeathstate
	end
end

SafeFreeslot("sfx_csnap", "sfx_shadsl", "sfx_shadsr", "sfx_habump",
	"sfx_damnit", "sfx_nothre", "sfx_damn4c", "sfx_ultvic", "sfx_shrail",
	"sfx_wspear", "sfx_mspear", "sfx_sprxpl", "sfx_sprcrg", "sfx_sprrdy",
	"sfx_chbrst", "sfx_chnght", "sfx_shadhm")

sfxinfo[sfx_csnap].caption = "Warp"
sfxinfo[sfx_shadsl].caption = "/"
sfxinfo[sfx_shadsr].caption = "/"
sfxinfo[sfx_habump].caption = "Bump"
sfxinfo[sfx_damnit].caption = "\"Damnit\""
sfxinfo[sfx_nothre].caption = "\"Damn, not here\""
sfxinfo[sfx_damn4c].caption = "\"Damn FOURTH Chaos Emerald\""
sfxinfo[sfx_ultvic].caption = "\"Ultimate victory!\""
sfxinfo[sfx_shrail].caption = "\"That rail sure came in handy\""
sfxinfo[sfx_wspear].caption = "Chaos Spear"
sfxinfo[sfx_mspear].caption = "Chaos Spear"
sfxinfo[sfx_sprxpl].caption = "Explosion"
sfxinfo[sfx_sprxpl].flags = $ | SF_X2AWAYSOUND
sfxinfo[sfx_sprcrg].caption = "Charging"
sfxinfo[sfx_sprrdy].caption = "Charged"
sfxinfo[sfx_chbrst].caption = "Chaos Burst"
sfxinfo[sfx_chnght].caption = "Chaos Nightmare"
sfxinfo[sfx_shadhm].caption = "Whoosh"

SafeFreeslot("SPR2_SHIT",
	"SPR2_WARP", "SPR2_POSE", "S_SHADOW_WARP1", "S_SHADOW_WARP2",
	"S_SHADOW_WARP_EXIT", "S_SHADOW_WARP_EXIT2",
	"S_SHADOW_BOUNCE", "S_SHADOW_POSE", "S_SHADOW_BOUNCE2",
	"S_SHADOW_SPEARCHARGE", "S_SHADOW_SPEARFIRE",
	"SPR2_BATB", "S_SHADOW_CHAOSBURST", "SPR2_BATN", "S_SHADOW_CHAOSNIGHTMARE")

spr2defaults[SPR2_SHIT] = SPR2_PAIN
spr2defaults[SPR2_POSE] = SPR2_SPNG

states[S_SHADOW_WARP1] = {
	sprite = SPR_NULL,
	tics = 8,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_SHADOW_WARP2
}

states[S_SHADOW_WARP2] = {
	sprite = SPR_PLAY,
	frame = SPR2_WARP | FF_SPR2ENDSTATE | FF_FULLBRIGHT,
	action = A_SetPAnim,
	tics = 1,
	nextstate = S_SHADOW_WARP2,
	var1 = S_SHADOW_WARP_EXIT,
	var2 = PA_ABILITY,
}

states[S_SHADOW_BOUNCE] = {
	sprite = SPR_PLAY,
	frame = SPR2_SPNG,
	tics = 16,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_SHADOW_BOUNCE2,
}

states[S_SHADOW_POSE] = {
	sprite = SPR_PLAY,
	frame = SPR2_POSE,
	tics = 16,
	action = A_SetPAnim,
	var2 = PA_ABILITY,
	nextstate = S_SHADOW_BOUNCE2,
}

states[S_SHADOW_BOUNCE2] = {
	sprite = SPR_PLAY,
	frame = SPR2_JUMP,
	tics = 0,
	action = A_ModifyPFlags,
	var2 = PF_NOJUMPDAMAGE,
	nextstate = S_PLAY_JUMP
}

states[S_SHADOW_WARP_EXIT] = {
	sprite = SPR_PLAY,
	frame = SPR2_FIRE,
	tics = 5,
	action = A_SetPAnim,
	var2 = PA_FALL,
	nextstate = S_SHADOW_WARP_EXIT2,
}

states[S_SHADOW_WARP_EXIT2] = {
	sprite = SPR_PLAY,
	frame = SPR2_FIRE,
	tics = 0,
	action = A_ModifyPFlags,
	var2 = PF_NOJUMPDAMAGE,
	nextstate = S_PLAY_JUMP,
}

states[S_SHADOW_SPEARCHARGE] = {
	sprite = SPR_PLAY,
	frame = SPR2_SPNG,
	tics = 2,
	action = A_ModifyPFlags,
	var1 = PF_NOJUMPDAMAGE,
	var2 = PF_SPINNING,
	nextstate = S_SHADOW_SPEARCHARGE,
}

states[S_SHADOW_SPEARFIRE] = {
	sprite = SPR_PLAY,
	frame = SPR2_FIRE,
	tics = 6,
	action = A_ModifyPFlags,
	var1 = PF_NOJUMPDAMAGE,
	var2 = PF_SPINNING,
	nextstate = S_PLAY_FALL,
}

SafeFreeslot("SPR2_SNAP", "S_CHAOSSNAP")

states[S_CHAOSSNAP] = {
	sprite = SPR_PLAY,
	frame = SPR2_SNAP | FF_SPR2ENDSTATE | FF_FULLBRIGHT,
	tics = 1,
	nextstate = S_CHAOSSNAP,
	var1 = S_NULL,
}

SafeFreeslot("MT_SHADOW_TRAIL", "MT_SHADOW_SKATEFLAME")

mobjinfo[MT_SHADOW_TRAIL] = {
	spawnstate = S_THOK,
	height = mobjinfo[MT_THOK].height,
	radius = mobjinfo[MT_THOK].radius,
	flags = mobjinfo[MT_THOK].flags
}

mobjinfo[MT_SHADOW_SKATEFLAME] = {
	spawnstate = S_INVISIBLE,
	height = mobjinfo[MT_PLAYER].height,
	radius = mobjinfo[MT_PLAYER].radius,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING|MF_NOGRAVITY|MF_SCENERY
}

SafeFreeslot("SPR_CSPP", "S_CHAOSSPEAR", "S_SPEAROFLIGHT", "S_CHAOSSPEAR_XPLD", "S_SPEAROFLIGHT_XPLD", "S_CHAOSSPEAR_SPARK", "MT_CHAOSSPEAR")

states[S_CHAOSSPEAR] = {
	sprite = SPR_CSPP,
	frame = FF_ANIMATE|FF_FULLBRIGHT,
	var1 = 2,
	var2 = 2,
	tics = 6,
	nextstate = S_CHAOSSPEAR,
}

states[S_SPEAROFLIGHT] = {
	sprite = SPR_CSPP,
	frame = J|FF_ANIMATE|FF_FULLBRIGHT,
	var1 = 2,
	var2 = 2,
	tics = 6,
	nextstate = S_SPEAROFLIGHT,
}

states[S_CHAOSSPEAR_XPLD] = {
	sprite = SPR_CSPP,
	frame = D|FF_ANIMATE|FF_FULLBRIGHT,
	var1 = 5,
	var2 = 2,
	tics = 12,
	action = A_SpearExplode
}

states[S_SPEAROFLIGHT_XPLD] = {
	sprite = SPR_CSPP,
	frame = M|FF_ANIMATE|FF_FULLBRIGHT,
	var1 = 5,
	var2 = 2,
	tics = 12
}

states[S_CHAOSSPEAR_SPARK] = {
	sprite = SPR_JETF,
	frame = D|FF_FULLBRIGHT,
	tics = 1,
	action = A_RollAngle,
	var1 = 20,
	nextstate = S_CHAOSSPEAR_SPARK,
}

mobjinfo[MT_CHAOSSPEAR] = {
	spawnstate = S_CHAOSSPEAR,
	raisestate = S_SPEAROFLIGHT,
	spawnhealth = 1000,
	seesound = sfx_wspear,
	attacksound = sfx_mspear,
	height = 24*FRACUNIT,
	radius = 25*FRACUNIT,
	damage = 128*FRACUNIT,
	speed = 80*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_MISSILE,
	deathstate = S_CHAOSSPEAR_XPLD,
	xdeathstate = S_SPEAROFLIGHT_XPLD,
	deathsound = sfx_sprxpl,
}

states[S_SHADOW_CHAOSBURST] = {
	sprite = SPR_PLAY,
	frame = SPR2_BATB,
	tics = 5,
	nextstate = S_SHADOW_CHAOSBURST,
}

SafeFreeslot("SPR_CBFW", "SPR_CBOR", "SPR_CBBK",
"MT_CHAOSBURST1", "MT_CHAOSBURST2", "MT_CHAOSBURST3",
"S_CHAOSBURST1", "S_CHAOSBURST2", "S_CHAOSBURST3")

local burstspeed = 3

states[S_CHAOSBURST1] = {
	sprite = SPR_CBFW,
	frame = FF_ANIMATE,
	var1 = 11,
	var2 = burstspeed,
	tics = burstspeed*12,
	nextstate = S_NULL,
}

states[S_CHAOSBURST2] = {
	sprite = SPR_CBOR,
	frame = FF_ANIMATE|FF_FULLBRIGHT|TR_TRANS30,
	var1 = 11,
	var2 = burstspeed,
	tics = burstspeed*12,
	nextstate = S_NULL,
}

states[S_CHAOSBURST3] = {
	sprite = SPR_CBBK,
	frame = FF_ANIMATE,
	var1 = 11,
	var2 = burstspeed,
	tics = burstspeed*12,
	nextstate = S_NULL,
}

mobjinfo[MT_CHAOSBURST1] = {
	spawnstate = S_CHAOSBURST1,
	spawnhealth = 1000,
	height = 1,
	radius = 1,
	speed = 1,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY,
	dispoffset = 6
}

mobjinfo[MT_CHAOSBURST2] = {
	spawnstate = S_CHAOSBURST2,
	spawnhealth = 1000,
	height = 1,
	radius = 1,
	speed = 1,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY,
	dispoffset = 5
}

mobjinfo[MT_CHAOSBURST3] = {
	spawnstate = S_CHAOSBURST3,
	spawnhealth = 1000,
	height = 1,
	radius = 1,
	speed = 1,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY,
	dispoffset = -1
}

states[S_SHADOW_CHAOSNIGHTMARE] = {
	sprite = SPR_PLAY,
	frame = SPR2_BATN,
	tics = 4,
	nextstate = S_SHADOW_CHAOSNIGHTMARE,
}

SafeFreeslot("SPR_CNMR",
"MT_CHAOSNIGHTMARE", "S_CHAOSNIGHTMARE")

states[S_CHAOSNIGHTMARE] = {
	sprite = SPR_CNMR,
	frame = FF_ANIMATE|FF_FULLBRIGHT|TR_TRANS30,
	var1 = 8,
	var2 = burstspeed,
	tics = burstspeed*9,
	nextstate = S_NULL,
	dispoffset = 5
}

mobjinfo[MT_CHAOSNIGHTMARE] = {
	spawnstate = S_CHAOSNIGHTMARE,
	spawnhealth = 1000,
	height = 1,
	radius = 1,
	speed = 1,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY,
}

SafeFreeslot("SPR_SOLC", "S_SPEAROFLIGHT_CHARGEORB")

states[S_SPEAROFLIGHT_CHARGEORB] = {
	sprite = SPR_SOLC,
	frame = FF_ANIMATE|FF_FULLBRIGHT|TR_TRANS30,
	var1 = 8,
	var2 = burstspeed,
	nextstate = S_NULL,
	dispoffset = 5
}