/*
Shadow
	by Lach
*/
/*
Some code for a specific secret
	by Rapidgame7
*/

addHook("PreThinkFrame", function()
	for p in players.iterate do
		pbtDaemon(p)
		p._sideforce = p.cmd.sidemove -- I need this because stasis sets this to zero afterwards
	end
end)

local SKIN = "shadow"

local CHAOS_SNAPS = 2 // number of chaos snaps per jump
local SNAP_HORIZONTAL_DIST = 72*8*FRACUNIT // horizontal distance to travel in a full chaos snap
local SNAP_VERTICAL_DIST = 25*8*FRACUNIT // vertical distance to travel in a full chaos snap
local SNAP_2D_DIST = 42*8*FRACUNIT // horizontal distance to travel in a full chaos snap, in 2d mode
local HOMING_FAIL_QUOTES = { // 1/25 chance one of these quotes will play when a homing attack fails
	sfx_damnit,
	sfx_nothre,
}
local HOMING_GHOST_TICS = 8 // amount of time to display thrusting homing attack ghosts
local HOMING_THRUST_SOUND = sfx_shadhm // sound to play when performing a thrusting homing attack
local LANCE_CHARGE_THRESHOLD = TICRATE/2 // amount of time to hold a chaos spear charge for Chaos Lance to occur
local BARRAGE_CHARGE_THRESHOLD = 2*TICRATE // amount of time to hold a chaos spear charge for Spear of Light to occur

local RUN_TICS = { // number of tics to display each running frame
	10,
	4,
	3,
	2,
	2,
	10,
	4,
	3,
	2,
	2,
}

local SKATE_FLAME_ANIMS = {
	[SPR2_RUN] = SPR2_TAL0|FF_ANIMATE, // NOTE: FF_ANIMATE signifies that the flames should match their frame with the player's instead of animating independently
	[SPR2_RUN|FF_SPR2SUPER] = SPR2_TAL0|FF_SPR2SUPER,
	[SPR2_SKID] = SPR2_TAL1|FF_SPR2ENDSTATE // NOTE: FF_SPR2ENDSTATE signifies that the flames should animate only once until the player leaves the spr2
}

local BUTTONS = {BT_USE}

// Shadow flags
local SHF_WARPING = 1 << 0 // Shadow is in a warp state; used to help toggle MF_NOGRAVITY
local SHF_JUSTWARPED = 1 << 1 // Shadow is teleporting; toggled off when the teleport ends so momentum can be halted
local SHF_WARPMASK = SHF_JUSTWARPED | SHF_WARPING
local SHF_SHIELDABILITY = 1 << 2 // Shadow used a shield ability and should not be able to use his main abilities
local SHF_SPEARCHARGE = 1 << 3 // Super Shadow is charging a chaos spear barrage

// Warptypes
local WARP_NONE = 0
local WARP_DOWN = 1
local WARP_HORIZONTAL = 2
local WARP_UP = 3

local cv_renderer = CV_FindVar("renderer")
local cv_chasecam = CV_FindVar("chasecam")
local cv_chasecam2 = CV_FindVar("chasecam2")

local function valid(mo)
	return mo and mo.valid
end

local function TwoD(mo)
	return twodlevel or mo.flags2 & MF2_TWOD
end

local spr2numframes = {}
local function GetSprite2NumFrames(skin, spr2)
	local frame
	
	if not spr2numframes[skin]
		spr2numframes[skin] = {}
	else
		frame = spr2numframes[skin][spr2]
	end
	
	if frame ~= nil
		return frame
	end
	
	local statenum = S_UNKNOWN
	local state = states[statenum]
	local statedef = {
		state.sprite,
		state.frame,
		state.tics,
		state.action,
		state.var1,
		state.var2,
		state.nextstate
	}
	local thok = P_SpawnMobj(0, 0, 0, MT_THOK)
	state.sprite = SPR_PLAY
	state.frame = spr2
	state.action = nil
	thok.skin = skin
	thok.state = statenum
	frame = -1
	while thok.frame & FF_FRAMEMASK > frame
		frame = thok.frame & FF_FRAMEMASK
		thok.state = statenum
	end
	frame = $ + 1
	spr2numframes[skin][spr2] = frame
	states[statenum] = statedef
	P_RemoveMobj(thok)
	return frame
end

local function ShieldVisibility(player, visible)
	if not (player.powers[pw_shield] & SH_NOSTACK)
	or player.powers[pw_super]
		return
	end
	
	local mo = player.mo
	
	for item in mobjs.iterate()
		local target = item
		for i = 1, 2
			if valid(target)
				if target.flags2 & MF2_SHIELD
				and target.target == mo
					if visible
						item.flags2 = $ & ~MF2_DONTDRAW
					else
						P_SpawnGhostMobj(item)
						item.flags2 = $ | MF2_DONTDRAW
					end
					break
				end
				target = target.target
			end
		end
	end
end

local function EndSpearCharge(player, allowbarrage)
	local mo = player.mo
	local shadow = player.shadow
	
	if not allowbarrage
		shadow.spearbarrage = 0
	end
	shadow.spearcharge = 0
	shadow.flags = $ & ~SHF_SPEARCHARGE
	S_StopSoundByID(mo, sfx_sprcrg)
end

local function FireChaosSpear(shadow, type)
	local missile
	local mo = shadow.mo
	
	type = $ or 0
	
	if type > 0
		local player = shadow.player
		local target = P_LookForEnemies(player)
		if valid(target)
			missile = P_SpawnMissile(mo, target, MT_CHAOSSPEAR)
			shadow.spearangle = R_PointToAngle2(mo.x, mo.y, target.x, target.y)
		else
			missile = P_SPMAngle(mo, MT_CHAOSSPEAR, mo.angle, 0)
			shadow.spearangle = mo.angle
		end
		
		if valid(missile)
			S_StopSoundByID(mo, missile.info.seesound)
			S_StartSound(mo, missile.info.attacksound)
			missile.flags2 = $ | MF2_AMBUSH | MF2_DONTRESPAWN
			missile.tracer = target
			if type == 2
				missile.flags2 = $ | MF2_SLIDEPUSH
				missile.state = missile.info.raisestate
				missile.color = SKINCOLOR_RUBY
			else
				missile.color = SKINCOLOR_WAVE
				missile.colorized = true
			end
		end
	else
		local missile = P_SPMAngle(mo, MT_CHAOSSPEAR, mo.angle, 0)
		if valid(missile)
			missile.flags2 = $ | MF2_DONTRESPAWN
			missile.color = SKINCOLOR_MINT
		end
		shadow.spearangle = mo.angle
	end
end

local function QueueSpearBarrage(shadow, spears)
	shadow.spearbarrage = (spears - 1)*4 + 1
	shadow.mo.tics = shadow.spearbarrage
end

local function DoHomingFail(mo)
	S_StartSound(mo, sfx_habump)
	if P_RandomChance(FRACUNIT/25)
		mo.state = S_PLAY_PAIN
		mo.player.pflags = $ | PF_NOJUMPDAMAGE
		S_StartSound(mo, HOMING_FAIL_QUOTES[P_RandomKey(#HOMING_FAIL_QUOTES) + 1])
	end
end

local function Unwarp(player)
	local mo = player.mo
	local shadow = player.shadow
	
	shadow.flags = $ & ~SHF_WARPMASK
	mo.flags = $ & ~MF_NOGRAVITY
	ShieldVisibility(player, true)
end

local function EndHomingAttack(shadow, bounceback)
	local mo = shadow.mo
	local player = shadow.player
	
	if TwoD(mo)
		mo.momy = 0
	end
	if mo.eflags & MFE_UNDERWATER
		P_SetObjectMomZ(mo, FixedDiv(457*FRACUNIT,72*FRACUNIT), false)
	else
		P_SetObjectMomZ(mo, 10*FRACUNIT, false)
	end
	if shadow.snaps < 1
		shadow.snaps = 1
	end
	if bounceback
		P_InstaThrust(mo, mo.angle, -(player.speed >> 3))
	else
		if player.nerfedhoming
			player.nerfedhoming = false
			//Let BattleMod handle the momentum with a nerfed homing attack
		elseif player.cmd.buttons & BT_JUMP
			P_InstaThrust(mo, mo.angle, FixedMul(player.actionspd / 6, mo.scale))
			S_StartSound(mo, HOMING_THRUST_SOUND)
			shadow.ghosttics = HOMING_GHOST_TICS
		else
			mo.momx = 0
			mo.momy = 0
		end
	end
	mo.state = S_SHADOW_BOUNCE
	for i = 1, shadow.homingchain % (GetSprite2NumFrames(SKIN, states[S_SHADOW_POSE].frame) + 1)
		mo.state = S_SHADOW_POSE
	end
	player.pflags = $ | PF_NOJUMPDAMAGE
	player.homing = 0
	shadow.homingchain = $ + 1
end

local function UpdateSkateFlame(player, flame)
	local mo = player.mo
	local data = mo and SKATE_FLAME_ANIMS[mo.sprite2]
	
	if not data
	or mo.sprite ~= SPR_PLAY
	or mo.eflags & MFE_UNDERWATER
		flame.sprite2 = SPR2_STND
		P_SetMobjStateNF(flame, flame.info.spawnstate)
		return
	end
	
	local frame = flame.frame & FF_FRAMEMASK
	local sprite2 = data & FF_FRAMEMASK
	local camangle
	if splitscreen
		camangle = player.cmd.angleturn << 16
	else
		camangle = R_PointToAngle(mo.x, mo.y)
	end
	
	flame.flags2 = $ & ~MF2_DONTDRAW | (mo.flags2 & MF2_DONTDRAW)
	flame.scale = mo.scale
	flame.color = player.shadow_skatecolor or player.powers[pw_super] and SKINCOLOR_SUPERSKATE or SKINCOLOR_SHADOWSKATE
	
	if flame.sprite2 ~= sprite2
		flame.skin = SKIN
		flame.sprite = SPR_PLAY
		flame.sprite2 = sprite2
		flame.flags2 = $ & ~(MF2_SLIDEPUSH|MF2_CLASSICPUSH|MF2_LINKDRAW)
		frame = -1
	end
	
	if data & FF_ANIMATE
		frame = mo.frame & FF_FRAMEMASK
		flame.flags2 = $ & ~(MF2_SLIDEPUSH|MF2_CLASSICPUSH)
	elseif not (flame.flags2 & MF2_CLASSICPUSH)
		flame.flags2 = $ ^^ MF2_SLIDEPUSH
		if flame.flags2 & MF2_SLIDEPUSH
			if data & FF_SPR2ENDSTATE
				frame = $ + 1
				if frame >= GetSprite2NumFrames(SKIN, sprite2)
					flame.flags2 = $ | MF2_CLASSICPUSH
					P_SetMobjStateNF(flame, flame.info.spawnstate)
				end
			else
				frame = ($ + 1) % GetSprite2NumFrames(SKIN, sprite2)
			end
		end
	end
	
	flame.frame = frame | FF_FULLBRIGHT | (leveltime & 1) * FF_TRANS40
	
	if cv_renderer.value == 2
		P_TeleportMove(flame,
			mo.x + P_ReturnThrustX(flame, camangle, -mo.scale),
			mo.y + P_ReturnThrustY(flame, camangle, -mo.scale),
			flame.z)
	end
end

local warpers = {}
local prevemeralds = 0

addHook("NetVars", function(net)
	warpers = net(warpers)
	prevemeralds = net(prevemeralds)
end)

addHook("ThinkFrame", do
	for player in players.iterate
		local mo = player.mo
		
		if not mo
			continue
		end
		
		if mo.skin ~= SKIN
		or player.playerstate
			if player.shadow
				player.shadow = nil
			end
			continue
		end
		
		if not player.shadow
		or player.shadow.mo ~= mo
			player.shadow = {}
			
			local shadow = player.shadow
			
			shadow.player = player
			shadow.warptype = WARP_NONE
			shadow.mo = mo
			shadow.flags = 0
			shadow.snaps = CHAOS_SNAPS
			shadow.prevwalkframe = -1
			shadow.prevrunframe = -1
			shadow.prevhoming = 0
			shadow.homingchain = 0
			shadow.time = -1
			shadow.bike = 0
			shadow.bikeprev = 0
			shadow.spearcharge = 0
			shadow.spearangle = 0
			shadow.bosshits = 0
			
			shadow.buttons = {}
			
			for i = 1, #BUTTONS
				shadow.buttons[BUTTONS[i]] = 0
			end
		end
		
		local shadow = player.shadow
		local grounded = P_IsObjectOnGround(mo)
		local flip = P_MobjFlip(mo)
		
		// count buttons held
		
		for i = 1, #BUTTONS
			shadow.buttons[BUTTONS[i]] = player.cmd.buttons & BUTTONS[i] and $ + 1 or 0
		end
		
		// count timer
		
		shadow.time = $ + 1
		
		// &: We don't want Shadow to be Shadow while on a bike
		if not shadow.bike then
			// floor hit stuff
			
			if not shadow.grounded and grounded
				if shadow.flags & SHF_WARPING // warped into a floor
				and not (player.pflags & PF_SPINNING)
					Unwarp(player)
					mo.state = S_SHADOW_WARP2
				end
				
				if TwoD(mo) and player.homing
					mo.momy = 0
				end
				
				if player.homing // cancel homing attack if hit a floor
				and not (player.pflags & PF_SHIELDABILITY)
				and not (mo.eflags & MFE_SPRUNG)
					player.homing = 0
					player.pflags = $ | PF_JUMPED | PF_THOKKED
					if shadow.snaps < 1
						shadow.snaps = 1
					end
					mo.state = S_PLAY_JUMP
					P_InstaThrust(mo, mo.angle, FixedMul(player.actionspd >> 3, mo.scale))
					P_SetObjectMomZ(mo, player.actionspd >> 4, false)
					if mo.eflags & MFE_UNDERWATER
						mo.momz = $ / 2
					end
					mo.z = $ + flip
					DoHomingFail(mo)
				else			
					player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE|PF_THOKKED)
					shadow.snaps = CHAOS_SNAPS
					shadow.flags = $ & ~(SHF_SHIELDABILITY)
					shadow.homingchain = 0
					shadow.bosshits = 0
					player.nerfedhoming = false
				end
			end
			
			if player.powers[pw_carry] and not shadow.carry
				shadow.snaps = CHAOS_SNAPS
				Unwarp(player)
			end
			
			// warp logic
			
			if mo.state == S_SHADOW_WARP1
				player.pflags = $ | PF_FULLSTASIS
			elseif mo.state == S_SHADOW_WARP2
			and shadow.flags & SHF_WARPING
				if shadow.flags & SHF_JUSTWARPED
					shadow.flags = $ & ~SHF_JUSTWARPED
					if mo.eflags & MFE_UNDERWATER
						mo.momx = $ / 4
						mo.momy = $ / 4
						if shadow.warptype ~= WARP_DOWN
							mo.momz = $ / 8
						end
					else
						mo.momx = $ / 2
						mo.momy = $ / 2
						if shadow.warptype ~= WARP_DOWN
							mo.momz = $ / 4
						end
					end
					shadow.warptype = WARP_NONE
					table.insert(warpers, mo)
					ShieldVisibility(player, true)
				end
			elseif shadow.flags & SHF_WARPING
				Unwarp(player)
			end
			
			// post-homing attack logic
			
			if shadow.prevhoming
			and not player.homing
			and not (mo.eflags & MFE_SPRUNG)
				// failed homing attack
				if player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE|PF_SPINNING|PF_SHIELDABILITY) == PF_JUMPED
					local maxmomz
					if mo.eflags & MFE_UNDERWATER
						maxmomz = FixedDiv(457*FRACUNIT,72*FRACUNIT)
					else
						maxmomz = 10*FRACUNIT
					end
					mo.momz = min(max($, -maxmomz), maxmomz)
					if TwoD(mo)
						mo.momy = 0
					end
					P_InstaThrust(mo,
						R_PointToAngle2(0, 0, mo.momx - player.cmomx, mo.momy - player.cmomy),
						FixedMul(player.actionspd >> 4, mo.scale))
				elseif mo.state == S_PLAY_SPRING // successful homing attack, but not on an enemy or monitor
					EndHomingAttack(shadow)
				end
			end
			
			if mo.eflags & MFE_SPRUNG // hit a vertical spring? thrust forward if you're holding jump
				if not (player.pflags & (PF_JUMPED|PF_THOKKED))
				and mo.momx == 0
				and mo.momy == 0
				and player.cmd.buttons & BT_JUMP
					P_InstaThrust(mo, mo.angle, shadow.prevspeed)
					S_StartSound(mo, HOMING_THRUST_SOUND)
					shadow.ghosttics = HOMING_GHOST_TICS
				end
			end
			
			if player.homing // cancel homing attack if too slow (hit a wall)
			and not (player.pflags & PF_SHIELDABILITY)
			and FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz) < FixedMul(player.actionspd, mo.scale) >> 2
				if TwoD(mo)
					mo.momy = 0
				end
				if shadow.snaps < 1
					shadow.snaps = 1
				end
				player.homing = 0
				P_InstaThrust(mo, mo.angle, -FixedMul(player.actionspd >> 4, mo.scale))
				P_SetObjectMomZ(mo, player.actionspd >> 4, false)
				if mo.eflags & MFE_UNDERWATER
					mo.momz = $ / 2
				end
				DoHomingFail(mo)
			end
			
			// Chaos Spears
			
			if player.powers[pw_super]
			and not (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6)
			and player.pflags & PF_JUMPED
			and not (shadow.flags & SHF_SHIELDABILITY)
			and not player.homing
				if shadow.flags & SHF_JUSTWARPED // don't allow spears to function while warping, but allow charging to continue
					EndSpearCharge(player, true)
				else
					if shadow.buttons[BT_USE] == 1 // pressed spin? start charging!
						shadow.flags = $ | SHF_SPEARCHARGE
						shadow.spearcharge = 0
					end
					
					if shadow.flags & SHF_SPEARCHARGE
					and not player.weapondelay
						if shadow.buttons[BT_USE] // charge...
							shadow.spearcharge = $ + 1
							if shadow.spearcharge == 6
								S_StartSound(mo, sfx_sprcrg)
								player.pflags = $ | PF_THOKKED
							elseif shadow.spearcharge >= LANCE_CHARGE_THRESHOLD
								if shadow.spearcharge == LANCE_CHARGE_THRESHOLD
									S_StartSound(mo, sfx_sprrdy)
									mo.state = S_SHADOW_SPEARCHARGE
								elseif shadow.spearcharge == BARRAGE_CHARGE_THRESHOLD
									S_StartSound(mo, sfx_sprrdy)
									Shadow_SpawnChaosNightmare(mo, S_SPEAROFLIGHT_CHARGEORB)
									mo.state = S_SHADOW_SPEARCHARGE
									mo.tics = -1
									mo.sprite2 = states[S_SHADOW_CHAOSNIGHTMARE].frame
									mo.frame = $ & ~FF_FRAMEMASK + GetSprite2NumFrames(mo.skin, states[S_SHADOW_CHAOSNIGHTMARE].frame) - 1
								end
								local chargecolor = shadow.spearcharge >= BARRAGE_CHARGE_THRESHOLD and SKINCOLOR_SUPERRED2 or SKINCOLOR_SUPERSKY1
								if shadow.spearcharge % 4 == 0
									local trail = P_SpawnGhostMobj(mo)
									trail.color = chargecolor
									trail.colorized = true
									trail.destscale = 0
									P_TeleportMove(trail, unpack(shadow.prevposition))
								end
								for i = 0, 1
									local x, y, z = Shadow_GetHandPositionFromAngle(mo, player.drawangle + ANGLE_90 + ANGLE_180*i)
									local spark = P_SpawnMobj(x, y, z + flip*16*mo.scale, MT_WATERZAP)
									local hangle = FixedAngle(360*P_RandomFixed())
									local vangle = FixedAngle(360*P_RandomFixed())
									local hdist = P_ReturnThrustX(spark, vangle, 6*mo.scale)
									if mo.eflags & MFE_VERTICALFLIP
										spark.flags2 = $ | MF2_OBJECTFLIP
										spark.eflags = $ | MFE_VERTICALFLIP
										spark.z = $ - spark.height
									end
									spark.scale = mo.scale
									spark.color = chargecolor
									spark.colorized = true
									spark.momx = mo.momx + P_ReturnThrustX(spark, hangle, hdist)
									spark.momy = mo.momy + P_ReturnThrustY(spark, hangle, hdist)
									spark.momz = mo.momz + P_ReturnThrustY(spark, vangle, 6*mo.scale)
								end
							end
						else // fire!
							mo.state = S_SHADOW_SPEARFIRE
							if shadow.spearcharge < LANCE_CHARGE_THRESHOLD // single spear
								FireChaosSpear(shadow)
							elseif shadow.spearcharge < BARRAGE_CHARGE_THRESHOLD // single homing spear (chaos lance)
								FireChaosSpear(shadow, 1)
							else // homing spear barrage (spear of light)
								QueueSpearBarrage(shadow, 9)
							end
							player.weapondelay = mo.tics
							EndSpearCharge(player, true)
						end
					end
					
					if shadow.spearbarrage
						shadow.spearbarrage = $ - 1
						if shadow.spearbarrage % 4 == 0
							FireChaosSpear(shadow, 2)
						end
					end
					
					if mo.state == S_SHADOW_SPEARCHARGE // float while charging/barraging
					or shadow.spearbarrage
						local potentialmomz = player.speed < 10*mo.scale and (player.speed - 10*mo.scale)/5 or 0
						if flip*mo.momz < potentialmomz
							mo.momz = flip*potentialmomz
						end
					end
					
					if mo.state == S_SHADOW_SPEARFIRE // face firing angle
						player.drawangle = shadow.spearangle
					end
				end
			elseif shadow.flags & SHF_SPEARCHARGE
			or shadow.spearbarrage
				EndSpearCharge(player)
			end
			
			// animation stuffs
			
			if mo.sprite2 == SPR2_WALK
				shadow.prevwalkframe = mo.frame & FF_FRAMEMASK
			end
			
			if mo.sprite2 == SPR2_RUN
				// lift a different foot depending on the previous walk frame
				// G H A B for left foot first
				// C D E F for right foot first
				if shadow.prevrunframe == -1
					local maxframe = GetSprite2NumFrames(SKIN, SPR2_WALK)
					if (shadow.prevwalkframe + 2) % maxframe < maxframe/2
						for i = 1, GetSprite2NumFrames(SKIN, SPR2_RUN)/2
							mo.state = $
						end
					end
				end
				
				// adjust timing and play sound effects
				local frame = mo.frame & FF_FRAMEMASK
				if frame ~= shadow.prevrunframe
					mo.tics = RUN_TICS[frame + 1] or 0
					if FixedDiv(player.speed, mo.movefactor) > 52*mo.scale
						mo.tics = max(1, $ - 1)
					end
					
					if grounded
						if frame == 0
							S_StartSound(mo, sfx_shadsr)
						elseif frame == GetSprite2NumFrames(SKIN, SPR2_RUN)/2
							S_StartSound(mo, sfx_shadsl)
						end
					end
					
					shadow.prevrunframe = frame
				end
			elseif shadow.prevrunframe >= 0
				shadow.prevrunframe = -1
			end

			// homing attack thrust ghosts
			
			if shadow.ghosttics
				shadow.ghosttics = $ - 1
				if shadow.time % 2
					local ghost = P_SpawnGhostMobj(mo)
					P_TeleportMove(ghost, unpack(shadow.prevposition))
					ghost.tics = 3
					ghost.colorized = true
					if not player.powers[pw_super]
					and player.skincolor == skins[mo.skin].prefcolor
						ghost.color = SKINCOLOR_SHADOWAURA
					end
				end
			end
			
			// Homing reticle
			
			if player.pflags & (PF_JUMPED|PF_THOKKED) == PF_JUMPED|PF_THOKKED
			and not (shadow.flags & (SHF_JUSTWARPED|SHF_SHIELDABILITY))
			and not player.homing
			and shadow.spearcharge < LANCE_CHARGE_THRESHOLD
				local target = P_LookForEnemies(player, true, false)
				if valid(target)
					P_SpawnLockOn(player, target, S_LOCKON1)
				end
			end
		end // &: End bike check
		
		// resets
		
		shadow.prevspeed = FixedHypot(mo.momx, mo.momy)
		shadow.prevposition = {mo.x, mo.y, mo.z}
		shadow.grounded = grounded
		shadow.carry = player.powers[pw_carry]
		shadow.prevhoming = player.homing
	end
	
	// warp effect ghosts
	
	for i = #warpers, 1, -1
		local mo = warpers[i]
		if not valid(mo)
		or mo.state ~= S_CHAOSSNAP and mo.state ~= S_SHADOW_WARP2
			table.remove(warpers, i)
			continue
		end
		
		local player = mo.player
		
		if netgame or modeattacking // always spawn ghosts in these modes to minimize desyncs
		or not player
		or player == displayplayer and cv_chasecam.value
		or player == secondarydisplayplayer and cv_chasecam2.value
			for i = 1, 2
				local ghost = P_SpawnGhostMobj(mo)
				local x = ghost.x
				local y = ghost.y
				local z = ghost.z
				local dist
				local hangle = FixedAngle(P_RandomKey(360) << FRACBITS)
				local vangle = FixedAngle(P_RandomRange(-90, 90) << FRACBITS)
				local hthrust = FixedMul(cos(vangle), ghost.scale)
				
				ghost.tics = 1
				
				if player == displayplayer and not cv_chasecam.value
					ghost.flags2 = $ | MF2_DONTDRAW
				end
				
				if mo.state == S_CHAOSSNAP // ghosts diverge
					dist = mo.frame & FF_FRAMEMASK
				elseif mo.state == S_SHADOW_WARP2 // ghosts converge
					dist = 6 - (mo.frame & FF_FRAMEMASK)
				end
				
				dist = max(0, $ - 2)
				dist = $ << 2
				
				x = $ + P_ReturnThrustX(mo, hangle, dist*hthrust)
				y = $ + P_ReturnThrustY(mo, hangle, dist*hthrust)
				z = $ + P_ReturnThrustY(mo, vangle, dist*mo.scale)
				
				P_TeleportMove(ghost, x, y, z)
			end
		end
	end
	
	// easter egg :)
	
	if emeralds & ~prevemeralds == EMERALD4
		local function IsShadow(player)
			return player and player.mo and player.mo.skin == SKIN
		end
		if IsShadow(displayplayer) or IsShadow(secondarydisplayplayer)
			S_StartSound(nil, sfx_damn4c)
		end
	end
	prevemeralds = emeralds
end)

addHook("PostThinkFrame", do
	for player in players.iterate
		local flame = player.followmobj
		if valid(flame) and flame.type == skins[SKIN].followitem
			UpdateSkateFlame(player, flame)
		end
	end
end)

// Ability hooks

addHook("AbilitySpecial", function(player)
	local mo = player.mo
	local shadow = player.shadow
	
	if not mo
	or mo.skin ~= SKIN
	or not shadow
	or player.charability ~= skins[SKIN].ability
	or player.gotflag
	or not (player.pflags & PF_JUMPED)
	or shadow.bike // &: HAHA
		return
	end
	
	if player.homing
	or shadow.flags & SHF_SHIELDABILITY and player.pflags & PF_THOKKED
	or shadow.spearcharge >= LANCE_CHARGE_THRESHOLD
		return true
	end
	
	local target = P_LookForEnemies(player, true, false)
	
	if target
		mo.tracer = target
		// Krabs
		if CBW_Battle and (target.type == MT_TARGETDUMMY or target.type == MT_SPARRINGDUMMY)
			//Shadow is homing on to a player in battlemod. Thats BROKEN better NERF THAT
			player.nerfedhoming = true
			player.homing = 10
			shadow.flags = $ | SHF_SHIELDABILITY //Prevent homing chains
		else
			player.nerfedhoming = false
			player.homing = 2*TICRATE
			shadow.flags = $ & ~SHF_SHIELDABILITY
		end
		// EndKrabs
		player.pflags = $ & ~PF_NOJUMPDAMAGE | PF_THOKKED
		mo.state = S_PLAY_JUMP
		S_StartSound(mo, sfx_thok)
		return true
	end
	
	if shadow.snaps <= 0
		return true
	end
	
	local intwod = TwoD(mo)
	
	S_StartSound(mo, sfx_csnap)
	
	local thok = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
	thok.skin = mo.skin
	thok.state = S_CHAOSSNAP
	thok.tics = $ + 1
	thok.color = mo.color
	thok.angle = player.drawangle
	table.insert(warpers, thok)
	
	player.pflags = $ & ~PF_SPINNING | PF_NOJUMPDAMAGE | PF_THOKKED
	
	shadow.flags = $ & ~SHF_SHIELDABILITY | SHF_WARPMASK
	if not player.powers[pw_super]
		shadow.snaps = $ - 1
	end
	
	mo.state = S_SHADOW_WARP1
	mo.flags = $ | MF_NOGRAVITY
	
	ShieldVisibility(player, false)
	
	mo.momx = 0
	mo.momy = 0
	mo.momz = 0
	
	if player.cmd.buttons & BT_USE
	or intwod and player.cmd.forwardmove < 0 and player.cmd.forwardmove < -abs(player.cmd.sidemove)
		shadow.warptype = WARP_DOWN
		P_SetObjectMomZ(mo, -SNAP_VERTICAL_DIST / mo.tics, false)
	elseif abs(player.cmd.sidemove) > abs(player.cmd.forwardmove)
	and intwod
		shadow.warptype = WARP_HORIZONTAL
		P_InstaThrust(mo, mo.angle, FixedMul(SNAP_2D_DIST / mo.tics, mo.scale))
	elseif (player.cmd.forwardmove or player.cmd.sidemove)
	and not intwod
		shadow.warptype = WARP_HORIZONTAL
		local dist = SNAP_HORIZONTAL_DIST * max(player.cmd.forwardmove, 25) / 50
		P_InstaThrust(mo,
			(player.cmd.angleturn << 16) + R_PointToAngle2(0, 0, player.cmd.forwardmove << 16, -player.cmd.sidemove << 16),
			FixedMul(dist / mo.tics, mo.scale))
	else
		shadow.warptype = WARP_UP
		P_SetObjectMomZ(mo, SNAP_VERTICAL_DIST / mo.tics, false)
	end
	
	return true
end)

local function HomingContact(item, mo)
	if valid(mo) and mo.player and mo.skin == SKIN
		local player = mo.player
		local shadow = player.shadow
		
		if player.homing
		and not (player.pflags & PF_SHIELDABILITY)
		and shadow
			if item.flags & MF_BOSS
				if shadow.bosstarget ~= item
					shadow.bosstarget = item
					shadow.bosshits = 0
				end
				shadow.bosshits = $ + 1
				if shadow.bosshits == item.info.spawnhealth
					S_StartSound(mo, sfx_ultvic)
				end
			end
			
			if item.type == MT_EGGGUARD and item.threshold ~= 42
			or item.type == MT_EGGSHIELD
			or item.health > 1
				if item.flags & MF_ENEMY
					P_InstaThrust(item, mo.angle + ANGLE_180, -(player.speed >> 1))
				end
				EndHomingAttack(shadow, true)
			else
				EndHomingAttack(shadow)
			end
		end
	end
end

addHook("MobjDamage", function(item, mo)
	if valid(item)
		return HomingContact(item, mo)
	end
end)

addHook("MobjDeath", function(item, mo)
	if valid(item) and item.flags & MF_MONITOR
		return HomingContact(item, mo)
	end
end)

addHook("ShouldDamage", function(mo, inflictor, source, damage, damagetype)
	if mo.valid and mo.skin == SKIN
	and mo.state >= S_SHADOW_WARP1
	and mo.state <= S_SHADOW_WARP2
		if valid(inflictor) and inflictor.flags & MF_BOXICON
		or damagetype & DMG_DEATHMASK
			return
		end
		return false
	end
end, MT_PLAYER)

addHook("ShieldSpecial", function(player)
	local mo = player.mo
	local shadow = player.shadow
	
	if mo and mo.skin == SKIN
	and shadow
		shadow.flags = $ | SHF_SHIELDABILITY
	end
end)

// Trail logic (some taken from source!)

addHook("MobjSpawn", function(mo)
	if valid(mo)
		mo.frame = FF_TRANS70
		mo.fuse = mo.tics
	end
end, MT_SHADOW_TRAIL)

addHook("MobjThinker", function(mo)
	if not valid(mo)
		return
	end
	
	if mo.flags2 & MF2_BOSSNOTRAP // "fast" flag
		if ((mo.frame & FF_TRANSMASK) >> FF_TRANSSHIFT) < (NUMTRANSMAPS-1) - (2*mo.fuse)/3
			// fade out when nearing the end of fuse...
			mo.frame = ($ & ~FF_TRANSMASK) | (((NUMTRANSMAPS-1) - (2*mo.fuse)/3) << FF_TRANSSHIFT)
		end
	else
		if ((mo.frame & FF_TRANSMASK) >> FF_TRANSSHIFT) < (NUMTRANSMAPS-1) - mo.fuse / 2
			// fade out when nearing the end of fuse...
			mo.frame = ($ & ~FF_TRANSMASK) | (((NUMTRANSMAPS-1) - mo.fuse / 2) << FF_TRANSSHIFT)
		end
	end
	
	if mo.flags2 & MF2_DONTRESPAWN
		return
	end
	
	mo.flags2 = $ | MF2_DONTRESPAWN
	
	local color = skins[SKIN].prefcolor
	
	// color object yellow if player is prefcolor
	if valid(mo.target)
		local player = mo.target.player
		if player
		and not player.powers[pw_super]
		and player.skincolor == color
			mo.color = SKINCOLOR_SHADOWAURA
		end
	end
end, MT_SHADOW_TRAIL)

// Chaos Spears

addHook("MobjThinker", function(mo)
	if not mo.valid
	or mo.flags & MF_NOCLIPTHING
		return
	end
	
	if mo.tics % 3 == 0 or mo.flags2 & MF2_AMBUSH // trails
		P_SpawnGhostMobj(mo).colorized = true
	end
	if mo.tics % 2
		local spark = P_SpawnMobjFromMobj(mo,
			P_RandomRange(-6, 6)*FRACUNIT,
			P_RandomRange(-6, 6)*FRACUNIT,
			P_RandomRange(0, 12)*FRACUNIT, MT_THOK)
		spark.state = S_CHAOSSPEAR_SPARK
		spark.target = mo
		spark.angle = mo.angle + ANGLE_90
		spark.color = mo.color
		spark.colorized = true
		spark.fuse = 8
		spark.destscale = 0
	end
	
	if mo.flags2 & MF2_AMBUSH // homing spear
		local tracer = mo.tracer
		if valid(tracer) and tracer.health // already directed towards an object
			return
		end
		local dist = FixedMul(RING_DIST, mo.scale)
		searchBlockmap("objects", function(mo, item) // look for an object to direct towards
			if item.valid
			and item.health
			and (not (item.flags & MF_SHOOTABLE and item.flags & (MF_MONITOR|MF_ENEMY|MF_BOSS))) ~= (not (item.flags2 & MF2_INVERTAIMABLE))
			and FixedHypot(FixedHypot(mo.x - item.x, mo.y - item.y), mo.z - item.z) <= dist
			and abs(R_PointToAngle2(mo.x, mo.y, item.x, item.y) - mo.angle) < ANGLE_90
			and P_CheckSight(mo, item)
				P_HomingAttack(mo, item)
				mo.tracer = item
				return true
			end
		end, mo, mo.x - dist, mo.x + dist, mo.y - dist, mo.y + dist)
	end
end, MT_CHAOSSPEAR)

addHook("ShouldDamage", function(item, spear, source, damage, damagetype)
	if valid(spear)
	and spear.type == MT_CHAOSSPEAR
	and spear.flags2 & MF2_SLIDEPUSH
	and damage == 1
		P_DamageMobj(item, spear, source, damage + 1, damagetype)
		return false
	end
end)

// funny easter egg

addHook("MobjDamage", function(mo, inflictor, source)
	if valid(inflictor) and inflictor.type == MT_REDRING and inflictor.flags2 & MF2_RAILRING
	and valid(source) and source.skin == SKIN
		S_StartSound(source, sfx_shrail)
	end
end, MT_PLAYER)

//addfile shadow1battle.pk3; addfile zba_battlemod-v6.1.pk3